<?php
namespace TrxEffects\Core\Modules\ImageEffects;

use TrxEffects\Plugin;
use TrxEffects\Core\Modules\Module;

if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly
}

class ImageEffectsElementor extends Module {

	/**
	 * Class constructor.
	 *
	 * Initializing module.
	 *
	 * @since 1.0.0
	 * @access public
	 */
	public function __construct() {
		add_action( 'trx_effects_action/init', array( $this, 'init' ) );
	}


	/**
	 * Init module
	 *
	 * Do init actions of the module
	 *
	 * @hook action 'trx_effects_action/init'
	 *
	 * @since 1.0.0
	 * @access public
	 */
	public function init() {

		// Add module-specific parameters to the params of the Elementor's module 'Image'
		add_action( 'elementor/element/before_section_end', array( $this, 'add_params_to_image' ), 10, 3 );

		// Add module-specific data-attributes to the elements layout
		add_action( 'elementor/frontend/element/before_render', array( $this, 'add_data_attributes_to_image' ), 10, 1 );
		add_action( 'elementor/frontend/widget/before_render', array( $this, 'add_data_attributes_to_image' ), 10, 1 );
	}


	/**
	 * Image effects for the module 'Image'
	 *
	 * Add module-specific parameters to the params for the Elementor's module 'Image'
	 *
	 * @hook action 'elementor/element/before_section_start'
	 * @hook action 'elementor/widget/before_section_start'
	 *
	 * @since 1.0.0
	 * @access public
	 */
	public function add_params_to_image( $element, $section_id, $args ) {

		if ( ! is_object( $element ) ) return;

		$el_name = $element->get_name();

		if ( 'image' == $el_name && 'section_image' === $section_id ) {

			$element->add_control( 'trx_effects_image_effects_heading', array(
									'type' => \Elementor\Controls_Manager::HEADING,
									'label' => esc_html__( 'Image effects', 'trx-effects' ),
									'separator' => 'before',
			) );

			$element->add_control( 'trx_effects_image_effect', array(
									'type' => \Elementor\Controls_Manager::SELECT,
									'label' => esc_html__("Effect on mouse hover", 'trx-effects'),
									'options' => apply_filters( 'trx_effects_filter/image_effects', array(
										''           => esc_html__( 'None', 'trx-effects' ),
										'on_ripple'  => esc_html__( 'Ripple', 'trx-effects' ),
										'on_ripple2' => esc_html__( 'Ripple 2', 'trx-effects' ),
										'on_waves'   => esc_html__( 'Waves', 'trx-effects' ),
										'on_waves2'  => esc_html__( 'Waves 2', 'trx-effects' ),
										'on_smudge'  => esc_html__( 'Smudge', 'trx-effects' ),
										'on_swap'    => esc_html__( 'Swap', 'trx-effects' ),
										'on_tint'    => esc_html__( 'Tint', 'trx-effects' ),
									) ),
									'prefix_class' => 'trx_effects_image_effects_',
									'default' => '',
			) );

			$element->add_control( 'trx_effects_image_effect_scale', array(
				'type' => \Elementor\Controls_Manager::SWITCHER,
				'label' => esc_html__("Scale on hover", 'trx-effects'),
				'label_on' => esc_html__( 'On', 'trx-effects' ),
				'label_off' => esc_html__( 'Off', 'trx-effects' ),
				'return_value' => '1',
				'default' => '',
				'condition' => array(
					'trx_effects_image_effect!' => '',
				),
			) );

			$element->add_control( 'trx_effects_image_effect_strength', array(
				'label' => esc_html__( 'Strength', 'trx-effects' ),
				'type' => \Elementor\Controls_Manager::SLIDER,
				'default' => array(
					'size' => 20,
					'unit' => 'px'
				),
				'range' => array(
					'px' => array(
						'min' => 5,
						'max' => 50
					),
				),
				'size_units' => array( 'px' ),
				'condition' => array(
					'trx_effects_image_effect' => array('on_waves', 'on_waves2', 'on_ripple', 'on_ripple2', 'on_smudge')
				),
			) );

			$element->add_control( 'trx_effects_image_effect_waves_factor', array(
				'label' => esc_html__( 'Waves frequency', 'trx-effects' ),
				'type' => \Elementor\Controls_Manager::SLIDER,
				'default' => array(
					'size' => 4,
					'unit' => 'px'
				),
				'range' => array(
					'px' => array(
						'min' => 1,
						'max' => 10
					),
				),
				'size_units' => array( 'px' ),
				'condition' => array(
					'trx_effects_image_effect' => array('on_waves', 'on_waves2')
				),
			) );

			$element->add_control( 'trx_effects_image_effect_waves_direction', array(
				'label' => esc_html__( 'Direction', 'trx-effects' ),
				'description' => esc_html__( 'Ripple direction: 0 - horizontal, 100 - vertical', 'trx-effects' ),
				'type' => \Elementor\Controls_Manager::SLIDER,
				'default' => array(
					'size' => 100,
					'unit' => 'px'
				),
				'range' => array(
					'px' => array(
						'min' => 0,
						'max' => 100
					),
				),
				'size_units' => array( 'px' ),
				'condition' => array(
					'trx_effects_image_effect' => array('on_ripple', 'on_ripple2')
				),
			) );

			$element->add_control( 'trx_effects_image_effect_tint_color', array(
				'label' => esc_html__( 'Tint color', 'trx-effects' ),
				'label_block' => false,
				'type' => \Elementor\Controls_Manager::COLOR,
				'global' => array(
					'active' => false,
				),
				'default' => '',
				'condition' => array(
					'trx_effects_image_effect' => 'on_tint',
				),
			) );

			$element->add_control( 'trx_effects_image_effect_swap_image', array(
				'type' => \Elementor\Controls_Manager::MEDIA,
				'label' => esc_html__( 'Swap image', 'trx-effects' ),
				'default' => array(
					'url' => '',
				),
				'condition' => array(
					'trx_effects_image_effect' => array('on_swap')
				),
			) );

			$element->add_control( 'trx_effects_image_effect_displacement', array(
				'type' => \Elementor\Controls_Manager::MEDIA,
				'label' => esc_html__( 'Displacement texture', 'trx-effects' ),
				'default' => array(
					'url' => '',
				),
				'condition' => array(
					'trx_effects_image_effect' => array('on_ripple2', 'on_swap')
				),
			) );

		}
	}


	/**
	 * Add "data-trx-effects-image-effects-*"  to the wrapper of the element
	 *
	 * Hook render elements and add data parameters to its layout
	 *
	 * @hook action 'elementor/frontend/element/before_render'
	 * @hook action 'elementor/frontend/widget/before_render'
	 *
	 * @since 1.0.0
	 * @access public
	 */
	public function add_data_attributes_to_image( $element ) {
		if ( is_object( $element ) ) {
			$el_name = $element->get_name();
			if ( 'image' == $el_name ) {
				$settings = $element->get_settings();
				if ( ! empty( $settings['trx_effects_image_effect'] ) ) {
					$data = array(
						'data-trx-effects-image-effect' => $settings['trx_effects_image_effect'],
						'data-trx-effects-image-effect-scale' => (int) $settings['trx_effects_image_effect_scale'],
						'data-trx-effects-image-effect-strength' => 60 - max( 5, min( 50, $settings['trx_effects_image_effect_strength']['size'] ) ),
					);
					if ( in_array( $settings['trx_effects_image_effect'], array('on_waves', 'on_waves2') ) ) {
						$data['data-trx-effects-image-effect-waves-factor'] = isset( $settings['trx_effects_image_effect_waves_factor']['size'] )
																		? max( 1, min( 10, $settings['trx_effects_image_effect_waves_factor']['size'] ) )
																		: 4;
					}
					if ( in_array( $settings['trx_effects_image_effect'], array('on_ripple', 'on_ripple2') ) ) {
						$data['data-trx-effects-image-effect-waves-direction'] = isset( $settings['trx_effects_image_effect_waves_direction']['size'] )
																		? $settings['trx_effects_image_effect_waves_direction']['size'] / 100
																		: 1;
					}
					if ( in_array( $settings['trx_effects_image_effect'], array('on_ripple2', 'on_swap') ) ) {
						$data['data-trx-effects-image-effect-displacement'] = $settings['trx_effects_image_effect_displacement']['url'];
						if ( $settings['trx_effects_image_effect'] == 'on_swap' ) {
							$data['data-trx-effects-image-effect-swap-image'] = $settings['trx_effects_image_effect_swap_image']['url'];
						}
					}
					if ( $settings['trx_effects_image_effect'] == 'on_tint' ) {
						$data['data-trx-effects-image-effect-tint-color'] = empty( $settings['trx_effects_image_effect_tint_color'] )
																	? apply_filters('trx_effects_filter_get_theme_accent_color', '#efa758')
																	: $settings['trx_effects_image_effect_tint_color'];
					}
					$element->add_render_attribute( '_wrapper', apply_filters( 'trx_effects_filter_image_effect_data', $data, $element ) );
				}
			}
		}
	}

}
