<?php
namespace TrxEffects\Core\Options;

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * Option field base class
 *
 * Abstract class for any options field
 *
 * @since 1.0.0
 */
abstract class CustomizerField {

	/**
	 * Customizer object.
	 *
	 * Holds the Customizer object
	 *
	 * @since 1.0.0
	 * @access protected
	 *
	 * @var customizer
	 */
	protected $customizer = null;

	/**
	 * Class constructor.
	 *
	 * Initializing field properties.
	 *
	 * @since 1.0.0
	 * @access public
	 */
	public function __construct( $customizer ) {
		$this->customizer = $customizer;
	}

	/**
	 * Get option.
	 *
	 * Return option value by name ( a shorthand for customizer->options->get_option() ).
	 *
	 * @since 1.0.0
	 * @access protected
	 */
	protected function get_option( $name, $defa = '', $strict_mode = true ) {
		return $this->customizer->options->get_option( $name, $defa, $strict_mode );
	}

	/**
	 * Check if option name exists
	 *
	 * Return true if an option with specified name exists.
	 * ( a shorthand for customizer->options->check_option() ).
	 *
	 * @since 1.0.0
	 * @access protected
	 */
	protected function check_option( $name ) {
		return $this->customizer->options->check_option( $name );
	}


	/**
	 * Put field value to the html tag attribute
	 *
	 * Convert complex value (array) to a string to put to the html tag attribute
	 *
	 * @param (number|string|array) field - an simple value or array with values
	 *
	 * @since 1.0.0
	 * @access public
	 */
	public function prepare_field_value( $field ) {
		if ( is_array( $field['val'] ) ) {
			$val = '';
			foreach ( $field['val'] as $k => $v ) {
				$val .= ( $val ? '|' : '' ) . $k . '=' . $v;
			}
		} else {
			$val = $field['val'];
		}
		return $val;
	}


	/**
	 * Display field layout.
	 *
	 * Display a single field layout.
	 *
	 * @param (string) name - a name of the field
	 * @param (array) field - an array with field description
	 * @param (string) group - an optional parameter with a name of the group
	 *
	 * @since 1.0.0
	 * @access public
	 */
	public function show_field( $name, $field, $group = '' ) {

		// Prepare 'name' for the group fields
		if ( ! empty( $group ) ) $name = $group . "[{$name}]";
		$id = str_replace( array( '[', ']' ), array( '_', '' ), $name );

		return ( ! empty( $field['class'] ) && strpos( $field['class'], 'trx_effects_new_row' ) !== false 
					? '<div class="trx_effects_new_row_before"></div>'
					: ''
					)
				. '<div class="trx_effects_options_item'
						. ' trx_effects_options_item_' . esc_attr( $field['type'] )
						. ( ! empty( $field['hidden'] ) && $field['type'] != 'hidden' ? ' trx_effects_options_item_hidden' : '' )
						. ( ! empty( $field['class'] ) ? ' ' . esc_attr( $field['class'] ) : '' )
					. '">'
						. '<h4 class="trx_effects_options_item_title'
							. ( ! empty( $field['title_class'] ) ? ' ' . esc_attr( $field['title_class'] ) : '' )
						. '">'
							. esc_html( $field['title'] )
						. '</h4>'
						. '<div class="trx_effects_options_item_data">'
							. '<div class="trx_effects_options_item_field' 
								. ( ! empty( $field['dir'] ) ? ' trx_effects_options_item_field_' . esc_attr( $field['dir'] ) : '' )
								. '"'
								. ' data-param="' . esc_attr( $name ) . '"'
							. '>'
								. $this->show_field_layout( $id, $name, $field )
							. '</div>'
							. ( ! empty( $field['override']['desc'] ) || ! empty( $field['desc'] )
								? ( '<div class="trx_effects_options_item_description">'
										. ( ! empty( $field['override']['desc'] ) 	// param 'desc' already processed with wp_kses()!
											? trim( $field['override']['desc'] ) 
											: trim( $field['desc'] )
											)
									. '</div>'
									)
								: ''
								)
						. '</div>'
					. '</div>';
	}


	/**
	 * Display an inner layout of the field.
	 *
	 * Display tags inside field's wrapper.
	 *
	 * @param (string) name - a name of the field
	 * @param (array) field - an array with field description
	 *
	 * @return (string) output - a string with a html layout of the field
	 *
	 * @since 1.0.0
	 * @access protected
	 */
	protected function show_field_layout( $id, $name, $field ) {
		return '';
	}

}
