<?php
namespace TrxEffects\Core\Utils;

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * File system utilities.
 *
 * Utilities for manipulation with files
 *
 * @since 1.0.0
 */
class Files {

	/**
	 * Class constructor.
	 *
	 * Initializing class properties.
	 *
	 * @since 1.0.0
	 * @access public
	 */
	public function __construct() {
		add_action( 'trx_effects_action/init', array( $this, 'init_filesystem' ), 0 );
	}


	/**
	 * Init WP Filesystem before the plugin init
	 *
	 * Do init actions of the module
	 *
	 * @hook action 'trx_effects_action/init' with priority 0
	 *
	 * @since 1.0.0
	 * @access public
	 */
	public function init_filesystem() {
        if ( ! function_exists( 'WP_Filesystem' ) ) {
            require_once trailingslashit( ABSPATH ) . 'wp-admin/includes/file.php';
        }
		if ( is_admin() ) {
			$url = admin_url();
			$creds = false;
			// First attempt to get credentials.
			if ( function_exists( 'request_filesystem_credentials' ) && false === ( $creds = request_filesystem_credentials( $url, '', false, false, array() ) ) ) {
				// If we comes here - we don't have credentials
				// so the request for them is displaying no need for further processing
				return false;
			}
	
			// Now we got some credentials - try to use them.
			if ( ! WP_Filesystem( $creds ) ) {
				// Incorrect connection data - ask for credentials again, now with error message.
				if ( function_exists( 'request_filesystem_credentials' ) ) request_filesystem_credentials( $url, '', true, false );
				return false;
			}
			
			return true; // Filesystem object successfully initiated.
		} else {
            WP_Filesystem();
		}
		return true;
	}


	/**
	 * Return path or url to the file/folder in the child or parent theme
	 *
	 * Check if file/folder present in the child theme and return path (url) to it.
	 * Else - return path (url) to file in the main theme dir (if present).
	 * Else - return path (url) to file in the plugin dir.
	 *
	 * @param (string) file - relative path to the file in the plugin directory
	 * @param (boolean) return_url - if true - URL is returned, else - local path
	 * 
	 * @return (string) return_dir - path or URL to the file
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function get_file_dir( $file, $return_url = false ) {
		if ( $file[0] == '/' ) $file = substr( $file, 1 );
		$theme_dir = get_template_directory() . '/' . TRX_EFFECTS_BASE . '/';
		$theme_url = get_template_directory_uri() . '/' . TRX_EFFECTS_BASE . '/';
		$child_dir = get_stylesheet_directory() . '/' . TRX_EFFECTS_BASE . '/';
		$child_url = get_stylesheet_directory_uri() . '/' . TRX_EFFECTS_BASE . '/';
		$dir = '';
		if ( file_exists( "{$child_dir}{$file}" ) ) {
			$dir = sprintf( '%1$s%2$s', $return_url ? $child_url : $child_dir, $file );
		} else if ( file_exists( "{$theme_dir}{$file}" ) ) {
			$dir = sprintf( '%1$s%2$s', $return_url ? $theme_url : $theme_dir, $file );
		} else if ( file_exists( sprintf( '%1$s%2$s', TRX_EFFECTS_DIR, $file ) ) ) {
			$dir = sprintf( '%1$s%2$s', $return_url ? TRX_EFFECTS_URL : TRX_EFFECTS_DIR, $file );
		}
		return apply_filters( $return_url ? 'trx_effects_filter/get_file_url' : 'trx_effects_filter/get_file_dir', $dir, $file );
	}


	/**
	 * Return url to the file/folder in the child or parent theme
	 *
	 * Check if file/folder present in the child theme and return url.
	 * Else - return url to file in the main theme dir (if present).
	 * Else - return url to file in the plugin dir.
	 *
	 * @param (string) file - relative path to the file in the plugin directory
	 * 
	 * @return (string) return_url - URL to the file
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function get_file_url( $file ) {
		return $this->get_file_dir( $file, true );
	}


	/**
	 * Return file extension
	 *
	 * Return file extension from full name/path
	 *
	 * @param (string) file - path to the file
	 * 
	 * @return (string) ext - extension of the file (if present) or empty string
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function get_file_ext( $file ) {
		$ext = pathinfo( $file, PATHINFO_EXTENSION );
		return empty( $ext ) ? '' : $ext;
	}


	/**
	 * Return file name
	 *
	 * Return file name from full name/path
	 *
	 * @param (string) file - path to the file
	 * @param (boolean) without_ext - return a file name without extension
	 * 
	 * @return (string) name - name of the file (if present) or empty string
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function get_file_name( $file, $without_ext = true ) {
		$parts = pathinfo( $file );
		return ! empty( $parts['filename'] ) && $without_ext ? $parts['filename'] : $parts['basename'];
	}


	/**
	 * Put data to the file
	 *
	 * Save data to the specified file (overwrite file or append data to the end of the file)
	 *
	 * @param (string) file - path to the file
	 * @param (string) data - data to save
	 * @param (int) flag - an operation type: 0 - overwrite, FILE_APPEND - append data
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function fpc( $file, $data, $flag = 0 ) {
		global $wp_filesystem;
		if ( ! empty( $file ) ) {
			if ( isset( $wp_filesystem ) && is_object( $wp_filesystem ) ) {
				$file = str_replace( ABSPATH, $wp_filesystem->abspath(), $file );
				// Attention! WP_Filesystem can't append the content to the file!
				if ( $flag == FILE_APPEND && $wp_filesystem->exists( $file ) && strpos( $file, '//' ) === false ) {
					// If it is a existing local file (not contain '//' in the path) and we need to append data -
					// use native PHP function to prevent large consumption of memory
					return file_put_contents( $file, $data, $flag );
				} else {
					// In other case (not a local file or not need to append data or file not exists)
					// That's why we have to read the contents of the file into a string,
					// add new content to this string and re-write it to the file if parameter $flag == FILE_APPEND!
					return $wp_filesystem->put_contents(
								$file, 
								( $flag == FILE_APPEND && $wp_filesystem->exists( $file )
									? $wp_filesystem->get_contents( $file )
									: ''
									)
									. $data,
								false
							);
				}
			} else {
				throw new Exception( sprintf( esc_html__( 'WP Filesystem is not initialized! Put contents to the file "%s" failed', 'trx-effects' ), $file ) );
			}
		}
		return false;
	}


	/**
	 * Get data from the file
	 *
	 * Return content of the specified file as string
	 *
	 * @param (string) file - path to the file
	 * @param (boolean) unpack - unpack data if a file is an archive
	 * 
	 * @return (string) content - a file content as a string
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function fgc( $file, $unpack = false ) {
		global $wp_filesystem;
		if ( ! empty( $file ) ) {
			if ( isset( $wp_filesystem ) && is_object( $wp_filesystem ) ) {
				$file = str_replace( ABSPATH, $wp_filesystem->abspath(), $file );
				$tmp_cont = strpos( $file, '//' ) !== false //&& ! $allow_url_fopen 
								? $this->remote_get( $file )
								: $wp_filesystem->get_contents( $file );
				if ( $unpack && $this->get_file_ext( $file ) == 'zip' ) {
					$tmp_name = 'tmp-' . rand() . '.zip';
					$tmp = wp_upload_bits( $tmp_name, null, $tmp_cont );
					if ( $tmp['error'] ) {
						$tmp_cont = '';
					} else {
						unzip_file( $tmp['file'], dirname( $tmp['file'] ) );
						$file_name = dirname( $tmp['file'] ) . '/' . basename( $file, '.zip' ) . '.txt';
						$tmp_cont = $this->fgc( $file_name );
						unlink( $tmp['file'] );
						unlink( $file_name );
					}
				}
				return $tmp_cont;
			} else {
				throw new Exception( sprintf( esc_html__( 'WP Filesystem is not initialized! Get contents from the file "%s" failed', 'trx-effects' ), $file ) );
			}
		}
		return '';
	}


	/**
	 * Get array of rows from the file
	 *
	 * Return content of the specified file as array of rows
	 *
	 * @param (string) file - path to the file
	 * 
	 * @return (string) content - a file content as an array
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function fga( $file ) {
		global $wp_filesystem;
		if ( ! empty( $file ) ) {
			if ( isset( $wp_filesystem ) && is_object( $wp_filesystem ) ) {
				$file = str_replace( ABSPATH, $wp_filesystem->abspath(), $file );
				return $wp_filesystem->get_contents_array( $file );
			} else {
				throw new Exception( sprintf( esc_html__( 'WP Filesystem is not initialized! Get rows from the file "%s" failed', 'trx-effects' ), $file ) );
			}
		}
		return array();
	}


	/**
	 * Remove unsafe characters from file/folder path
	 *
	 * Return a file path (url) without dangerous characters
	 *
	 * @param (string) name - path to the file
	 * 
	 * @return (string) name - a cleared file name
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function esc( $name ) {
		return str_replace( array( '\\', '~', '$', ':', ';', '+', '>', '<', '|', '"', "'", '`', "\xFF", "\x0A", "\x0D", '*', '?', '^' ), '/', trim( $name ) );
	}


	/**
	 * Get a file mode
	 *
	 * Return a file mode (file permissions supported by chmod)
	 *
	 * @param (string) filename - path to the file
	 * 
	 * @return (string) val - a file mode
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function getmod( $filename ) {
		$val = 0;
		$perms = fileperms( $filename );
		// Owner; User
		$val += ( ( $perms & 0x0100 ) ? 0x0100 : 0x0000 ); //Read
		$val += ( ( $perms & 0x0080 ) ? 0x0080 : 0x0000 ); //Write
		$val += ( ( $perms & 0x0040 ) ? 0x0040 : 0x0000 ); //Execute

		// Group
		$val += ( ( $perms & 0x0020 ) ? 0x0020 : 0x0000 ); //Read
		$val += ( ( $perms & 0x0010 ) ? 0x0010 : 0x0000 ); //Write
		$val += ( ( $perms & 0x0008 ) ? 0x0008 : 0x0000 ); //Execute

		// Global World
		$val += ( ( $perms & 0x0004 ) ? 0x0004 : 0x0000 ); //Read
		$val += ( ( $perms & 0x0002 ) ? 0x0002 : 0x0000 ); //Write
		$val += ( ( $perms & 0x0001 ) ? 0x0001 : 0x0000 ); //Execute

		// Misc
		$val += ( ( $perms & 0x40000 ) ? 0x40000 : 0x0000 ); //temporary file (01000000)
		$val += ( ( $perms & 0x80000 ) ? 0x80000 : 0x0000 ); //compressed file (02000000)
		$val += ( ( $perms & 0x100000 ) ? 0x100000 : 0x0000 ); //sparse file (04000000)
		$val += ( ( $perms & 0x0800 ) ? 0x0800 : 0x0000 ); //Hidden file (setuid bit) (04000)
		$val += ( ( $perms & 0x0400 ) ? 0x0400 : 0x0000 ); //System file (setgid bit) (02000)
		$val += ( ( $perms & 0x0200 ) ? 0x0200 : 0x0000 ); //Archive bit (sticky bit) (01000)

		return $val;
	}


	/**
	 * Get a remote file content via HTTP method GET
	 *
	 * Get text from specified file via HTTP method GET (cURL)
	 *
	 * @param (string) file - URL of the file
	 * @param (int) timeout - max connection timeout (in seconds). -1 - get timeout from php.ini
	 * 
	 * @return (string) response - a file content or empty string
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function remote_get( $file, $timeout = -1 ) {
		// Set timeout as half of the PHP execution time
		if ( $timeout < 1 ) $timeout = round( 0.5 * max( 30, ini_get( 'max_execution_time' ) ) );
		if ( substr( $file, 0, 2 ) == '//' ) $file = 'http' . ( is_ssl() ? 's' : '' ) . ':' . $file;
		$response = wp_remote_get( $file, array(
									'method'  => 'GET',
									'timeout' => $timeout
									)
								);
		//return wp_remote_retrieve_response_code( $response ) == 200 ? wp_remote_retrieve_body( $response ) : '';
		return ! is_wp_error( $response ) && isset( $response['response']['code'] ) && $response['response']['code'] == 200
					? $response['body']
					: '';
	}


	/**
	 * Get a remote file content via HTTP method POST
	 *
	 * Get text from specified file via HTTP method POST (cURL)
	 *
	 * @param (string) file - URL of the file
	 * @param (array) args - array with argumets to send to the server
	 * @param (int) timeout - max connection timeout (in seconds). -1 - get timeout from php.ini
	 * 
	 * @return (string) response - a file content or empty string
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function remote_post( $file, $args, $timeout = -1 ) {
		// Set timeout as half of the PHP execution time
		if ( $timeout < 1 ) $timeout = round( 0.5 * max( 30, ini_get( 'max_execution_time' ) ) );
		if ( substr( $file, 0, 2 ) == '//') $file = 'http' . ( is_ssl() ? 's' : '' ) . ':' . $file;
		$response = wp_remote_post( $file, array(
									'method'  => 'POST',
									'timeout' => $timeout,
									'body'    => $args
									)
								);
		//return wp_remote_retrieve_response_code( $response ) == 200 ? wp_remote_retrieve_body( $response ) : '';
		return ! is_wp_error( $response ) && isset( $response['response']['code'] ) && $response['response']['code'] == 200
					? $response['body']
					: '';
	}


	/**
	 * Get JSON from a remote file url via HTTP (cURL)
	 *
	 * Get JSON from specified url via HTTP (cURL) and return object or null
	 * 
	 * @param (string) url - URL of the file
	 * 
	 * @return (object) data - JSON data from the file
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function retrieve_json( $url ) {
		$data = '';
		$resp = trim( $this->remote_get( $url ) );
		if ( in_array( substr( $resp, 0, 1 ), array( '{', '[' ) ) ) {
			$data = json_decode( $resp, true );
		}
		return $data;
	}


	/**
	 * Delete a folder's tree
	 *
	 * Delete specified folder and it's subfolders
	 * 
	 * @param (string) dir - path to the folder
	 * @param (boolean) delete_root - true, if root folder also must be deleted
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function del_folder( $dir, $delete_root = true ) {
		$files = array();
		$dirs = array();
		$d = opendir( $dir );
		while ( ( $fname = readdir( $d ) ) !== false ) {
			if ( $fname == '.' || $fname == '..' ) {
				continue;
			}
			if ( is_dir( $dir . '/' . $fname ) ) {
				$dirs[] = $fname;
			} else {
				$files[] = $fname;
			}
		}
		closedir( $d );
		for ( $i = 0; $i < count( $files ); $i++ ) {
			unlink( $dir . '/' . $files[ $i ] );
		}
		for ( $i = 0; $i < count( $dirs ); $i++ ) {
			$this->del_folder( $dir . '/' . $dirs[ $i ] );
		}
		if ( $delete_root ) {
			rmdir( $dir );
		}
	}


	/**
	 * Get files list from the folder
	 *
	 * Get list of the files in the specified folder
	 * 
	 * @param (string) dir - path to the folder
	 * @param (string) relative_dir - (internal) dir to continue scan
	 * @param (array) allow_ext - return only files with specified extensions
	 * 
	 * @return (array) list - array (list) with files
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function scan_folder( $dir, $relative_dir = '', $allow_ext = array() ) {
		$list = array();
		$full_dir = $relative_dir . ( ! empty( $relative_dir ) && ! empty( $dir ) ? '/' : '') . $dir;
		if ( is_dir( $full_dir ) ) {
			$hdir = opendir( $full_dir );
			if ( $hdir ) {
				while ( ( $file = readdir( $hdir ) ) !== false ) {
					$pi = pathinfo( $full_dir . '/' . $file );
					if ( substr( $file, 0, 1) == '.' ) {
						continue;
					}
					if ( is_dir( $full_dir . '/' . $file ) ) {
						$list = array_merge( $list, $this->scan_folder( ( ! empty( $dir ) ? $dir . '/' : '') . $file, $relative_dir, $allow_ext ) );
					} else if ( empty( $allow_ext ) || in_array( $pi['extension'], $allow_ext ) ) {
						$list[] = ( ! empty( $dir ) ? $dir . '/' : '' ) . $file;
					}
				}
				closedir( $hdir );
			}
		}
		return $list;
	}


	/**
	 * Unpack archive
	 *
	 * Unpack archive to the specified folder
	 * 
	 * @param (string) arh_name - path to the archive
	 * @param (string) dest_folder - path to the destination folder
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function unpack_archive( $arh_name, $dest_folder ) {
		if ( class_exists( 'ZipArchive' ) ) {
			$zip = new ZipArchive;
			if ( $zip->open( $arh_name ) === true ) {
				$zip->extractTo( $dest_folder );
				// Restore date of each file
				$numFiles = $zip->numFiles;
				for ( $i=0; $i < $numFiles; $i++ ) {
					$info = $zip->statIndex( $i );
					$file = $dest_folder . '/' . $info['name'];
					if ( file_exists( $file ) || is_dir( $file ) ) {
						touch( $file, $info['mtime'] );
					}
				}
				$zip->close();
			}
		}
	}


	/**
	 * Pack archive
	 *
	 * Pack files from the specified folder to the archive
	 * 
	 * @param (string) arh_name - path to the archive
	 * @param (string) dest_folder - path to the destination folder
	 * @param (boolean|array) files - array with file paths to put in the archive. If specified - dest_folder is ignored
	 * 
	 * @return (boolean) rez - true if archive created successfully
	 * 
	 * @since 1.0.0
	 * @access public
	 */
	public function pack_archive( $arh_name, $dest_folder, $files=false ) {
		$rez = false;
		if ( class_exists( 'ZipArchive' ) ) {
			$zip = new ZipArchive;
			if ( $zip->open( $arh_name, ZIPARCHIVE::CREATE ) === true ) {
				if ( is_array( $files) ) {										// Add to archive files from the list
					foreach( $files as $file ) {
						$zip->addFile( $file, str_replace( $dest_folder . '/', '', $file ) );
					}
				} else if ( is_dir( $dest_folder ) ) {							// Add to archive specified folder
					$files = $this->scan_folder( '', $dest_folder );
					foreach( $files as $file ) {
						$zip->addFile( $dest_folder . '/' . $file, $file );
					}
				} else {														// Add to archive single file
					$pi = pathinfo( $dest_folder );
					$zip->addFile( $dest_folder, $pi['basename'] );
				}
				$zip->close();
				$rez = true;
			}
		}
		return $rez;
	}

}
